package net.locusworks.test;

import net.locusworks.common.annotations.MapValue;
import net.locusworks.common.utils.Utils;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.junit.jupiter.params.ParameterizedTest;

import java.lang.reflect.InvocationTargetException;
import java.util.*;
import java.util.stream.Stream;

import static org.junit.jupiter.api.Assertions.*;

/**
 * Test cases for the Utils class
 * @author Isaac Parenteau
 * @since 1.0.0-RELEASE
 *
 */
public class UtilsTest {

  @Test
  public void testSafeString() {
    assertNotNull(Utils.safeString(null));
    assertTrue(Utils.safeString(null).isEmpty());
    assertFalse(Utils.safeString("hello world").isEmpty());
  }


  @Test
  public void testAreEqual() {
    String val1 = "H";
    String val2 = "H";
    assertTrue(Utils.areEqual(val1, val2));

    assertThrows(IllegalArgumentException.class, () -> Utils.areEqual("1"));
  }
  @ParameterizedTest
  @MethodSource("areEqualParams")
  public void testAreEqualWithParams(List<Object> objectList, boolean or, boolean equal) {
    assertEquals(equal, Utils.areEqual(or, objectList.toArray()));
  }

  static Stream<Arguments> areEqualParams() {
    return Stream.of(
      Arguments.of(List.of("Hello", "Hello"), false, true),
      Arguments.of(List.of("Hello", "World"), false, false),
      Arguments.of(List.of("Hello", "World", "Hello"), true, true),
      Arguments.of(List.of("Hello", "World", "Fair"), true, false)
    );
  }
  
  @Test
  public void testEmptyString() {
    assertTrue(Utils.isEmptyString(null));
    assertTrue(Utils.isEmptyString(""));
    assertTrue(Utils.isEmptyString(" "));
    assertFalse(Utils.isEmptyString("foo"));
    assertFalse(Utils.isEmptyString("   bar   "));
  }
  
  @Test
  public void testToInteger() {
    assertEquals(2, (int) Utils.toInteger("Hello word", 2));
    assertEquals(23, (int) Utils.toInteger("23", 5023));
  }

  @ParameterizedTest
  @MethodSource("validateValueParams")
  public <V> void testValidateValue(V value, boolean valid) {
    assertEquals(valid, Utils.validateValue(value));
  }

  static Stream<Arguments> validateValueParams() {
    return Stream.of(
      Arguments.of(null, false),
      Arguments.of(Collections.emptyList(), false),
      Arguments.of(Map.of(), false),
      Arguments.of(false, false),
      Arguments.of("", false),
      Arguments.of("Hello", true),
      Arguments.of(List.of("Hello"), true),
      Arguments.of(Map.of("Hello", "World"), true),
      Arguments.of(true, true)
    );
  }

  @ParameterizedTest
  @MethodSource("areValidParams")
  public void testValidateValues(List<Object> objects, boolean or, boolean valid) {
      assertEquals(valid, Utils.areValid(objects.toArray()));
      assertEquals(!valid, Utils.areNotValid(objects.toArray()));
  }

  static Stream<Arguments> areValidParams() {
    return Stream.of(
      Arguments.of(List.of("Hello", List.of("Hello"), Map.of("Hello", "World"), true), false, true),
      Arguments.of(List.of(Collections.emptyList(), Map.of(), false), false, false)
    );
  }

  @Test
  public void testBuildMap() {
    Map<String, String> mymap = Utils.buildMap(TreeMap.class, String.class, String.class, "Hello", "World");
    assertNotNull(mymap);
    assertEquals(1, mymap.size());
    assertEquals("World", mymap.get("Hello"));
  }

  @Test
  public void testCloneList() {
    String[] values = new String[]{
      "Hello",
      "world"
    };
    List<String> cloned = Utils.cloneList(new ArrayList<>(Arrays.stream(values).toList()));
    assertEquals(2, cloned.size());
    assertThrows(UnsupportedOperationException.class, () -> cloned.add("asdf"));
  }

  @Test
  public void testCloneObject()
    throws InvocationTargetException, InstantiationException, IllegalAccessException,
    NoSuchMethodException {
    TestClass clazz = new TestClass();
    clazz.field1 = "hi";
    clazz.field2 = "low";
    clazz.field3 = "world";

    TestClass clazz3 = new TestClass();
    clazz3.field1 = "hi";
    clazz3.field2 = "smash";
    clazz3.field3 = "world";

    clazz.testClass = clazz3;

    TestClass clazz2 = Utils.cloneObject(clazz);
    assertNotNull(clazz2);
    assertEquals(clazz.field1, clazz2.field1);
    assertEquals(clazz.field2, clazz2.field2);
    assertEquals(clazz.field3, clazz2.field3);
    assertEquals(clazz.field5, clazz2.field5);
    assertEquals(clazz.number, clazz2.number);
    assertEquals(clazz.aBoolean, clazz2.aBoolean);
    assertNotNull(clazz2.field4);
    assertNull(clazz2.nullField);
  }

  @Test
  public void testCovertToMap() throws Exception {
    TestClass clazz = new TestClass();
    clazz.field1 = "hi";
    clazz.field2 = "low";
    clazz.field3 = "world";

    TestClass clazz3 = new TestClass();
    clazz.field1 = "hi";
    clazz.field2 = "low";
    clazz.field3 = "world";
    clazz.testClass = clazz3;

    Map<String, Object> converted = Utils.convertToMap(clazz);

    assertEquals(7, converted.size());
    assertEquals("hi", converted.get("other_field"));
  }

  @Test
  public void testConvertToStringMap() throws Exception {
    TestClass clazz = new TestClass();
    clazz.field1 = "hi";
    clazz.field2 = "low";
    clazz.field3 = "world";
    Map<String, String> converted = Utils.convertToStringMap(clazz);
    assertEquals(6, converted.size());
    assertEquals("hi", converted.get("other_field"));

    converted = Utils.convertToStringMap(converted);
    assertEquals(6, converted.size());
    assertEquals("hi", converted.get("other_field"));
  }

  @Test
  public void testBuildStringMap() {
    Map<String, String> myMap = Utils.buildStringHashMap("hello", "world");
    assertNotNull(myMap);
    assertEquals(1, myMap.size());
  }

  @Test
  public void testBuildMapExceptions() {
    Throwable ex = assertThrows(IllegalArgumentException.class, () -> Utils.buildMap(String.class, String.class, String.class, "" ));

    assertNotNull(ex);

    ex = assertThrows(IllegalArgumentException.class, () -> Utils.buildMap(TreeMap.class, String.class, String.class, "Hello"));

    assertEquals("Odd number of arguments provided", ex.getMessage());

    assertThrows(IllegalArgumentException.class, () -> Utils.buildMap(TreeMap.class, String.class, String.class, new Object(), "Hello"));

    assertThrows(IllegalArgumentException.class, () -> Utils.buildMap(TreeMap.class, String.class, String.class,  "Hello", new Object()));

    assertThrows(IllegalArgumentException.class, () -> Utils.buildMap(TreeMap.class, Object.class, String.class,  "Hello", new Object()));
  }

  @Test
  public void testBuildSet() {
    Set<String> myset = Utils.buildSet("Hello", "World");
    assertNotNull(myset);
    assertEquals(2, myset.size());
  }

  @Test
  public void testIsJunitRunning() {
    assertTrue(Utils.isJUnitRunning());
  }

  public static class TestClass {
    @MapValue("other_field")
    private String field1;

    @MapValue(ignore = true)
    private String field2;


    private String field3;

    private final String field4 = "final";
    @MapValue
    private int field5 = 5;

    private Boolean aBoolean = Boolean.valueOf("true");

    private Double number = 1.0d;

    private String nullField = null;

    private TestClass testClass;
  }

}
